import numpy as np
from sklearn.decomposition import IncrementalPCA, PCA


def fit(x, components: int = None, incremental: bool = False,
        batch_size: int = None, roi=None):
    """Calculate the first N principal components transforms for :code:`x`.

    :param x: Multichannel image of shape :code:`(C, H, W)`.
    :param components: Number of components to compute. Must be in the range
           :code:`1 < components <= C`.
    :param incremental: If :code:`True`, use Incremental PCA. Faster for large
           datasets at the expense of a slight decrease in accuracy.
    :param batch_size: Size of the input batch when using Incremental PCA.
    :param roi: If provided, only fit to the given region-of-interest.
    :return: The fitted PCA instance.
    """
    # Validate number of components
    if components is not None and not (1 < components <= x.shape[0]):
        raise ValueError(f'Requested components ({components}) outside '
                         f'range [1, {x.shape[0]}]')

    # Setup new PCA
    if incremental:
        pca = IncrementalPCA(n_components=components,
                             batch_size=batch_size)
    else:
        pca = PCA(n_components=components)

    # Crop training data to ROI
    if roi is not None:
        x = x[:, roi.y:roi.y + roi.h, roi.x:roi.x + roi.w]

    # Flatten input
    x_flat = x.reshape((x.shape[0], -1))
    x_flat = np.swapaxes(x_flat, 0, 1)

    # Fit input files
    pca.fit(x_flat)
    return pca


def apply_transform(x, pca) -> np.ndarray:
    """Apply precomputed PCA transforms to a multichannel image.

    :param x: Multichannel input image.
    :param pca: PCA instance returned by :func:`~educelab.imgproc.pca.fit`.
    :return: The transformed image.
    """
    # Transform images
    x_flat = x.reshape((x.shape[0], -1))
    x_flat = np.swapaxes(x_flat, 0, 1)
    x_flat = pca.transform(x_flat)

    # Convert back to images
    x_flat = np.swapaxes(x_flat, 0, 1)
    pca_shape = (pca.n_components_,) + x.shape[1:]
    return x_flat.reshape(pca_shape)



import numpy as np


def exposure(image, val):
    """Increase image exposure.

    :param image: Input image.
    :param val: Exposure adjustment factor (+/-).
    :return: Exposure adjusted image.
    """
    return image * 2 ** val


def shadows(image, val):
    """Image shadow adjustment.

    Adapted from an implementation by
    `HViktorTsoi <https://gist.github.com/HViktorTsoi/8e8b0468a9fb07842669aa368382a7df>`_.

    :param image: Input image.
    :param val: Shadow adjustment factor (+/-).
    :return: Exposure adjusted image.
    """
    shadow_val = 1. + val / 100. * 2
    shadow_mid = 3. / 10.
    shadow_region = np.clip(1. - image / shadow_mid, 0, 1)
    shadow_region[np.where(image >= shadow_mid)] = 0
    return (1 - shadow_region) * image + shadow_region * (
            1 - np.power(1 - image, shadow_val))



import subprocess
from pathlib import Path
from typing import List


def write(file: Path, tags: List[str]):
    """Write image metadata tags to an image file using ExifTool. Existing file
    will be overwritten.

    :param file: Path to image file.
    :param tags: List of ExifTool tag assignment strings.
    """
    # Setup metadata copy
    cmd = ['exiftool', '-q', '-overwrite_original', str(file)]

    # Map all the other tags
    cmd.extend(tags)

    # Run exiftool
    subprocess.run(cmd, check=True)


def copy_all(src: Path, dest: Path, extra_tags: List[str] = None,
             overwrite=True):
    """Copy tag metadata from source image to destination image.

    :param src: Path to image file containing source metadata.
    :param dest: Path to destination image file.
    :param extra_tags: If provided, write additional tags to :code:`dest`.
    :param overwrite: If :code:`True`, overwrite the destination image file.
    """
    # Setup metadata copy
    cmd = ['exiftool', '-q']

    # Overwrite original destination file
    if overwrite:
        cmd.append('-overwrite_original')

    # Skip tags that don't make sense in JPGs
    if dest.suffix not in ['.tif', '.tiff']:
        cmd.extend(['-XMP-tiff:all=', '-ExifIFD:BitsPerSample=',
                    '-IFD0:BitsPerSample='])

    # Original tags from original files
    cmd.extend(['-TagsFromFile', str(src)])

    # Map all the other tags
    cmd.append('-all:all')

    # Project specific tags
    if extra_tags is not None:
        cmd.extend(extra_tags)

    # Add target
    cmd.append(str(dest))

    # Run exiftool
    subprocess.run(cmd, check=True)



from functools import partial

import numpy as np
from skimage import exposure


def flatfield_correction(image, lf, df):
    """Apply `flatfield correction <https://en.wikipedia.org/wiki/Flat-field_correction>`_
    to an image.

    :param image: Input image.
    :param lf: Light-field image.
    :param df: Dark-field image.
    :return: Flatfield-corrected image.
    """
    fd_diff = lf - df
    return (image - df) * np.mean(fd_diff) / fd_diff


gamma_correction = exposure.adjust_gamma
"""Apply gamma correction to an image.

See: `skimage.exposure.adjust_gamma() <https://scikit-image.org/docs/stable/api/skimage.exposure.html#skimage.exposure.adjust_gamma>`_
"""

normalize = partial(exposure.rescale_intensity, out_range=(0., 1.))
"""Rescale an image's effective dynamic range (min and max values) to the 
range [0, 1]."""



from .adjust import (exposure, shadows)
from .conversion import (as_dtype)
from .correction import (flatfield_correction, gamma_correction, normalize)
from .enhance import (clahe, clip, curves, stretch, stretch_percentile, stretch_binned_percentile)
from .filters import (sharpen)



from collections.abc import Sequence
from typing import Union

import numpy as np
from skimage import exposure
from scipy.interpolate import CubicSpline

List2F = Sequence[float, float, ...]

clahe = exposure.equalize_adapthist
"""Contrast limited adaptive histogram equalization (CLAHE).

See: `skimage.exposure.equalize_adapthist() <https://scikit-image.org/docs/stable/api/skimage.exposure.html#skimage.exposure.equalize_adapthist>`_
"""

clip = np.clip
"""Clip image values to a range."""


def curves(image, x, y=None, **kwargs):
    """Cubic spline curves enhancement, aka Photoshop Curves.

    Constructs and applies a cubic spline intensity transfer function to the
    input image.

    :param image: Input image
    :param x: If shape (n,), the spline knot positions in the input intensity
           domain (a.k.a. the independent variables). If shape (n, 2), then
           a list of pairs defining the spline knot positions in both the input
           and output domains.
    :param y: If provided, an array_like of shape (n,) corresponding to spline
           knot positions in the output intensity range (a.k.a. the dependent
           variables). Required if x has shape (n,).
    :param kwargs: Extra kwargs passed to
           `scipy.interpolate.CubicSpline() <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.CubicSpline.html>`_.
    :return: The enhanced image.
    """
    if not isinstance(x, np.ndarray):
        x = np.array(x)

    if x.ndim == 0 or x.ndim > 2:
        raise ValueError('x must be of shape (n,) or (n, 2)')

    if y is None:
        if x.shape[-1] != 2:
            raise ValueError('missing y of shape (n,) or x of shape (n, 2)')
        y = x[..., 1]
        x = x[..., 0]
    else:
        y = np.ndarray(y)
        if x.ndim != 1 or y.ndim != 1:
            raise ValueError('x and y must be of shape (n,)')
        elif x.shape != y.shape:
            raise ValueError('x and y do not have the same length n')

    # default kwargs
    defaults = {'bc_type': 'natural'}
    kwargs = defaults | kwargs

    return CubicSpline(x, y, **kwargs)(image)


def stretch(image, a_min, a_max, **kwargs):
    """Linearly rescale the provided min/max range to the dynamic range of the
    image.
    """
    return exposure.rescale_intensity(image, in_range=(a_min, a_max), **kwargs)


def stretch_percentile(image, min_perc, max_perc, **kwargs):
    """Linearly rescale the provided percentile range to the dynamic range of
    the image.
    """
    min_val = np.percentile(image, min_perc)
    max_val = np.percentile(image, max_perc)
    return exposure.rescale_intensity(image, in_range=(min_val, max_val), **kwargs)


def stretch_binned_percentile(image, percent: Union[float, List2F, None] = None,
                              bins=256, **kwargs):
    """Alternative version of
    :func:`~educelab.imgproc.enhance.stretch_percentile` that uses percentiles
    calculated from intensity binning.
    """
    if percent is None:
        percent = (.35, .35)
    elif isinstance(percent, float):
        percent = (percent, percent)
    elif isinstance(percent, tuple):
        pass
    else:
        raise ValueError(
            f'unsupported type {type(percent)}, must be '
            f'[float, Sequence[float, float], None]')

    # calculate histogram
    hist, edges = np.histogram(image, bins=bins)

    # find the lower and upper bins which saturate clip_% pixels low and high
    threshold = int(image.size * percent[0] / 200.)
    c = (np.cumsum(hist) < threshold).argmin()
    threshold = int(image.size * percent[1] / 200.)
    d = bins - 1 - (np.cumsum(hist[::-1]) < threshold).argmin()

    # convert the bin to a low and high pixel value
    c = edges[0] + c * (edges[1] - edges[0])
    d = edges[0] + d * (edges[1] - edges[0])

    # rescale and return
    return exposure.rescale_intensity(image, in_range=(c, d), **kwargs)
